"""Separate environment-centric module"""

import typing as t

from named_env import (
    EnvironmentNamespace,
    OptionalString,
    OptionalTernary,
    OptionalBoolean,
    OptionalList,
)

__all__ = [
    "Env",
]


class Env(EnvironmentNamespace):
    """
    GRANA_LOG_LEVEL:
        Specifies the log level.
        Default is ERROR.
    GRANA_LOG_FILE:
        Specifies the log file.
        Defaults to the standard error stream.
    GRANA_ENV_FILE:
        Which file to load environment variables from. Expected format is k=v.
        Default is .env in the current directory.
    GRANA_WORKFLOW_FILE:
        Workflow file to use.
        Default behaviour is scan the current working directory.
    GRANA_WORKFLOW_LOADER_SOURCE_FILE:
        May point a file containing a WorkflowLoader class definition, which will replace the default implementation.
    GRANA_DISPLAY_NAME:
        Select the display by name from the bundled list.
    GRANA_DISPLAY_SOURCE_FILE:
        May point a file containing a Display class definition, which will replace the default implementation.
    GRANA_STRATEGY_NAME:
        Specifies the execution strategy.
        Default is 'loose'.
    GRANA_FORCE_COLOR:
        When specified, this will force the colored or non-coloured output, according to the setting.
    GRANA_SHELL_INJECT_YIELD_FUNCTION:
        When set to True, all shell-related actions will inject the yield_outcome function definition.
        Default is True.
    GRANA_EXTERNAL_MODULES_PATHS:
        A comma-separated list of local directories, which are added to the sys.path while loading any external modules.
        Default is an empty list.
    GRANA_ACTIONS_CLASS_DEFINITIONS_DIRECTORY:
        A comma-separated list of local directories, from which all `*.py` files will be considered action definitions.
        Each loaded definition is named after the filename stem and must contain an `Action` class.
        e.g. foo-bar.py may be referenced in a YAML workflow as `type: foo-bar`.
    GRANA_STRICT_OUTCOMES_RENDERING:
        When set to True, rendering a missing outcome key will result in an error instead of an empty string.
        Default is False.
    GRANA_DEFAULT_SHELL_EXECUTABLE:
        Specifies which shell executable should be used by the shell action by default.
        Default is /bin/sh.
    """

    GRANA_LOG_LEVEL: str = OptionalString("")
    GRANA_LOG_FILE: str = OptionalString("")
    GRANA_ENV_FILE: str = OptionalString("")
    GRANA_WORKFLOW_FILE: str = OptionalString("")
    GRANA_WORKFLOW_LOADER_SOURCE_FILE: str = OptionalString("")
    GRANA_DISPLAY_NAME: str = OptionalString("")
    GRANA_DISPLAY_SOURCE_FILE: str = OptionalString("")
    GRANA_STRATEGY_NAME: str = OptionalString("")
    GRANA_FORCE_COLOR: t.Optional[bool] = OptionalTernary(None)  # type: ignore
    GRANA_SHELL_INJECT_YIELD_FUNCTION: bool = OptionalBoolean(True)  # type: ignore
    GRANA_EXTERNAL_MODULES_PATHS: t.List[str] = OptionalList([])
    GRANA_ACTIONS_CLASS_DEFINITIONS_DIRECTORY: t.List[str] = OptionalList([])
    GRANA_STRICT_OUTCOMES_RENDERING: bool = OptionalBoolean(True)  # type: ignore
    GRANA_DEFAULT_SHELL_EXECUTABLE: str = OptionalString("/bin/sh")



"""Command-line interface entry"""

import functools
import os
import sys
import typing as t
from pathlib import Path

import classlogging
import click
from dotenv.main import DotEnv

from .config.constants import C, LOG_LEVELS
from .config.constants.cli import cliargs_receiver
from .config.environment import Env
from .display.default import KNOWN_DISPLAYS
from .exceptions import BaseError, ExecutionFailed
from .runner import Runner
from .strategy import KNOWN_STRATEGIES
from .tools.proxy import DeferredCallsProxy
from .version import __version__

logger = DeferredCallsProxy(obj=classlogging.get_module_logger())


# pylint: disable=invalid-name
class _WorkflowPositionalArgument(click.Argument):
    """Optional positional argument for the workflow source"""

    NAME: str = "WORKFLOW"

    # pylint: disable=unused-argument
    def __init__(self, param_decls: t.Sequence[str], required: t.Optional[bool] = None, **attrs: t.Any) -> None:
        super().__init__(param_decls, required=False, nargs=-1)

    def get_help_record(self, ctx: click.Context) -> t.Optional[t.Tuple[str, str]]:
        return self.make_metavar(), (
            "Workflow source file. When not given, will look for one of grana.yml/grana.yaml "
            "files in the context directory. Use the '-' value to read yaml configuration from the standard input. "
            "Also configurable via the GRANA_WORKFLOW_FILE environment variable."
        )

    def process_value(self, ctx: click.Context, value: t.Any) -> t.Optional[str]:
        """Check that there is not more than one argument for the workflow source"""
        vanilla_value: t.Tuple[str, ...] = super().process_value(ctx, value)
        if len(vanilla_value) > 1:
            raise click.BadParameter("Cannot apply more than one value", param=self)
        return vanilla_value[0] if vanilla_value else None

    def make_metavar(self) -> str:
        """Fixed representation"""
        return f"[{self.NAME}]"


@click.group
@click.option(
    "-l",
    "--log-level",
    help="Logging level. Defaults to ERROR. Also configurable via the GRANA_LOG_LEVEL environment variable.",
    type=click.Choice(list(LOG_LEVELS)),
)
@click.option(
    "-d",
    "--display",
    help="Display name. Defaults to prefixes. Also configurable via the GRANA_DISPLAY_NAME environment variable.",
    type=click.Choice(list(KNOWN_DISPLAYS)),
)
@cliargs_receiver
def main() -> None:
    """Declarative task runner"""


def load_dotenv() -> None:  # pragma: no cover
    """Try loading environment from the dotenv file.
    Special variable called "HERE" is injected into the environment during dotenv loading,
    which points to the directory of the dotenv file (if not specified in advance)."""
    here_var_name: str = "HERE"
    here_value_was_defined: bool = here_var_name in os.environ
    dotenv_path: Path = C.ENV_FILE
    if not here_value_was_defined:
        os.environ[here_var_name] = str(dotenv_path.parent)
    else:
        logger.debug(f"{here_var_name!r} was set externally")
    try:
        dotenv = DotEnv(dotenv_path=dotenv_path)
        if here_var_name in dotenv.dict():
            logger.debug(f"{here_var_name!r} is explicitly set via dotenv file")
            here_value_was_defined = True
        if dotenv.set_as_environment_variables():
            logger.info(f"Loaded environment variables from {str(dotenv_path)!r}")
        else:
            logger.debug(f"Dotenv not found: {str(dotenv_path)!r}")

    finally:
        if not here_value_was_defined:
            os.environ.pop(here_var_name)


def wrap_cli_command(func):
    """Standard loading and error handling"""

    @main.command
    @cliargs_receiver
    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        load_dotenv()
        classlogging.configure_logging(
            level=C.LOG_LEVEL,
            colorize=C.USE_COLOR and not C.LOG_FILE,
            main_file=C.LOG_FILE,
            stream=None if C.LOG_FILE else classlogging.LogStream.STDERR,
        )
        logger.uncork()
        try:
            return func(*args, **kwargs)
        except BaseError as e:
            logger.debug("", exc_info=True)
            sys.stderr.write(f"! {e}\n")
            sys.exit(e.CODE)
        except ExecutionFailed:
            logger.debug("Some steps failed")
            sys.exit(1)
        except Exception as e:
            logger.debug("", exc_info=True)
            sys.stderr.write(f"! UNHANDLED EXCEPTION: {e!r}\n")
            sys.exit(2)

    return wrapped


@wrap_cli_command
@click.option(
    "-s",
    "--strategy",
    help="Execution strategy. Defaults to loose. Also configurable via the GRANA_STRATEGY_NAME environment variable.",
    type=click.Choice(list(KNOWN_STRATEGIES)),
)
@click.option("-i", "--interactive", help="Run in dialog mode.", is_flag=True, default=False)
@click.argument("workflow", cls=_WorkflowPositionalArgument)
def run() -> None:
    """Run pipeline immediately."""
    Runner().run_sync()


@wrap_cli_command
@click.argument("workflow", cls=_WorkflowPositionalArgument)
def validate() -> None:
    """Check workflow validity."""
    action_num: int = len(Runner().workflow)
    logger.info(f"Located actions number: {action_num}")


@main.group
def info() -> None:
    """Package information."""


@info.command
def version() -> None:
    """Show package version."""
    print(__version__)


@info.command
def env_vars() -> None:
    """Show environment variables that are taken into account."""
    print(Env.__doc__)



"""YAML-based workflow load routines"""

from __future__ import annotations

import typing as t
from functools import lru_cache
from pathlib import Path

import yaml

from .base import AbstractBaseWorkflowLoader
from ..actions.base import ActionBase
from ..actions.bundled import (
    EchoAction,
    ShellAction,
    SubflowAction,
    DockerShellAction,
)
from ..actions.types import ObjectTemplate, Import
from ..config.constants import C
from ..config.constants.helpers import maybe_class_from_module
from ..exceptions import YAMLStructureError

__all__ = [
    "DefaultYAMLWorkflowLoader",
]


class YAMLLoader(yaml.SafeLoader):
    """Extension loader"""

    @classmethod
    def add_string_constructor(cls, tag: str, target_class: type) -> None:
        """Register simple string constructor with type checking"""

        def construct(_, node):
            if not isinstance(node.value, str):
                raise YAMLStructureError(f"Expected string content after {tag!r}, got {node.value!r}")
            return target_class(node.value)

        cls.add_constructor(tag, construct)


YAMLLoader.add_string_constructor("!@", ObjectTemplate)
YAMLLoader.add_string_constructor("!import", Import)


class DefaultYAMLWorkflowLoader(AbstractBaseWorkflowLoader):
    """Default loader for YAML source files"""

    ALLOWED_ROOT_TAGS: t.Set[str] = {"actions", "context", "miscellaneous", "configuration"}
    STATIC_ACTION_FACTORIES = {
        name: klass
        for name, klass in (
            ("echo", EchoAction),
            ("shell", ShellAction),
            ("subflow", SubflowAction),
            ("docker-shell", DockerShellAction),
        )
        if klass is not None
    }

    def _get_action_factory_by_type(self, action_type: str) -> t.Type[ActionBase]:
        if (dynamically_resolved_action_class := self._load_external_action_factories().get(action_type)) is not None:
            return dynamically_resolved_action_class
        return super()._get_action_factory_by_type(action_type)

    @lru_cache(maxsize=1)
    def _load_external_action_factories(self) -> t.Dict[str, t.Type[ActionBase]]:
        dynamic_bases_map: t.Dict[str, t.Type[ActionBase]] = {}
        for class_directory in C.ACTION_CLASSES_DIRECTORIES:  # type: str
            class_directory_path = Path(class_directory).resolve()
            self.logger.info(f"Loading external action classes from {str(class_directory_path)!r}")
            for class_file in class_directory_path.iterdir():
                if not class_file.is_file() or not class_file.suffix == ".py":
                    continue
                action_type: str = class_file.stem
                self.logger.debug(f"Trying external action class source: {class_file}")
                action_class: t.Type[ActionBase] = t.cast(
                    t.Type[ActionBase],
                    maybe_class_from_module(
                        path_str=str(class_file),
                        class_name="Action",
                        submodule_name=f"actions.{action_type}",
                    ),
                )
                if action_type in dynamic_bases_map:
                    self.logger.warning(f"Class {action_type!r} is already defined: overriding from {class_file}")
                dynamic_bases_map[action_type] = action_class
        return dynamic_bases_map

    def _parse_import(self, tag: Import, allowed_root_keys: t.Set[str]) -> None:
        path: str = tag.path
        if not path:
            self._throw(f"Empty import: {path!r}")
        with self._read_file(path) as file_data:
            self._internal_loads_with_filter(
                data=file_data,
                allowed_root_keys=allowed_root_keys,
            )

    def _internal_loads(self, data: t.Union[str, bytes]) -> None:
        self._internal_loads_with_filter(
            data=data,
            allowed_root_keys=self.ALLOWED_ROOT_TAGS,
        )

    def _internal_loads_with_filter(
        self,
        data: t.Union[str, bytes],
        allowed_root_keys: t.Set[str],
    ) -> None:
        if isinstance(data, bytes):
            data = data.decode()
        root_node: dict = yaml.load(data, YAMLLoader)  # nosec
        if not isinstance(root_node, dict):
            self._throw(f"Unknown workflow structure: {type(root_node)!r} (should be a dict)")
        root_keys: t.Set[str] = set(root_node)
        if not root_keys:
            self._throw(f"Empty root dictionary (expected some of: {', '.join(sorted(self.ALLOWED_ROOT_TAGS))}")
        if unrecognized_keys := root_keys - self.ALLOWED_ROOT_TAGS:
            self._throw(
                f"Unrecognized root keys: {sorted(unrecognized_keys)} "
                f"(expected some of: {', '.join(sorted(self.ALLOWED_ROOT_TAGS))}"
            )
        processable_keys: t.Set[str] = set(root_node) & allowed_root_keys
        if "actions" in processable_keys:
            actions: t.List[t.Union[dict, Import]] = root_node["actions"]
            if not isinstance(actions, list):
                self._throw(f"'actions' contents should be a list (got {type(actions)!r})")
            for child_node in actions:
                if isinstance(child_node, dict):
                    action: ActionBase = self.build_action_from_dict_data(child_node)
                    self._register_action(action)
                elif isinstance(child_node, Import):
                    self._parse_import(
                        tag=child_node,
                        allowed_root_keys={"actions"},
                    )
                else:
                    self._throw(f"Unrecognized node type: {type(child_node)!r}")
        if "context" in processable_keys:
            context: t.Union[t.Dict[str, str], t.List[t.Union[t.Dict[str, str], Import]]] = root_node["context"]
            if isinstance(context, dict):
                self._loads_contexts_dict(data=context)
            elif isinstance(context, list):
                for num, item in enumerate(context):
                    if isinstance(item, dict):
                        self._loads_contexts_dict(data=item)
                    elif isinstance(item, Import):
                        self._parse_import(
                            tag=item,
                            allowed_root_keys={"context"},
                        )
                    else:
                        self._throw(f"Context item #{num + 1} is not a dict nor an '!import' (got {type(item)!r})")
            else:
                self._throw(f"'context' contents should be a dict or a list (got {type(context)!r})")
        if "configuration" in processable_keys:
            self.load_configuration_from_dict(root_node["configuration"])

    def _loads_contexts_dict(self, data: t.Dict[str, t.Any]) -> None:
        for context_key, context_value in data.items():
            if not isinstance(context_key, str):
                self._throw(f"Context keys should be strings (got {type(context_key)!r} for {context_key!r})")
            if context_key in self._gathered_context:
                self.logger.debug(f"Context key redefined: {context_key}")
            else:
                self.logger.debug(f"Context key added: {context_key}")
            self._gathered_context[context_key] = context_value



# pylint: disable=import-outside-toplevel,cyclic-import
"""Lazy-loaded constants"""

import os
import sys
import typing as t
from io import UnsupportedOperation
from pathlib import Path

from classlogging import LogLevel

from .cli import get_cli_arg
from .helpers import (
    Optional,
    Mandatory,
    maybe_path,
    maybe_class_from_module,
)
from ..environment import Env
from ...types import (
    LoaderClassType,
    StrategyClassType,
    DisplayClassType,
)

__all__ = [
    "C",
    "LOG_LEVELS",
]

LOG_LEVELS: t.Dict[str, str] = {
    "0": LogLevel.ERROR,
    "1": LogLevel.WARNING,
    "2": LogLevel.INFO,
    "3": LogLevel.DEBUG,
    "4": LogLevel.TRACE,
    LogLevel.ERROR: LogLevel.ERROR,
    LogLevel.WARNING: LogLevel.WARNING,
    LogLevel.INFO: LogLevel.INFO,
    LogLevel.DEBUG: LogLevel.DEBUG,
    LogLevel.TRACE: LogLevel.TRACE,
}


def _maybe_strategy(name: t.Optional[str]) -> t.Optional[StrategyClassType]:
    """Transform an optional strategy name into an optional strategy class"""
    from ...strategy import KNOWN_STRATEGIES

    try:
        return KNOWN_STRATEGIES[name] if name else None
    except KeyError:
        raise ValueError(f"Invalid strategy name: {name!r} (allowed: {sorted(KNOWN_STRATEGIES)})") from None


def _maybe_display_class_by_name(name: t.Optional[str]) -> t.Optional[DisplayClassType]:
    from ...display.default import KNOWN_DISPLAYS

    if not name:
        return None
    try:
        return KNOWN_DISPLAYS[name]
    except Exception:
        raise ValueError(f"Display name should be one of: {sorted(KNOWN_DISPLAYS)}. Got {name!r}") from None


def _get_default_display_class() -> DisplayClassType:
    from ...display.default import DefaultDisplay

    return DefaultDisplay


def _get_strategy_class_from_cli_arg() -> t.Optional[StrategyClassType]:
    from ...strategy import KNOWN_STRATEGIES

    return _maybe_strategy(get_cli_arg("strategy", valid_options=KNOWN_STRATEGIES))


def _get_default_strategy_class() -> StrategyClassType:
    from ...strategy import LooseStrategy

    return LooseStrategy


def _isatty() -> bool:
    try:
        return os.isatty(sys.stdout.fileno())
    except UnsupportedOperation:
        return False


class C:
    """Runtime constants"""

    LOG_LEVEL: Mandatory[str] = Mandatory(
        lambda: LOG_LEVELS[get_cli_arg("log_level")] if get_cli_arg("log_level") is not None else None,
        lambda: Env.GRANA_LOG_LEVEL or None,
        lambda: LogLevel.ERROR,
    )
    LOG_FILE: Optional[Path] = Optional(
        lambda: maybe_path(Env.GRANA_LOG_FILE),
    )
    ENV_FILE: Mandatory[Path] = Mandatory(
        lambda: maybe_path(Env.GRANA_ENV_FILE),
        lambda: Path().resolve() / ".env",
    )
    CONTEXT_DIRECTORY: Mandatory[Path] = Mandatory(
        lambda: Path().resolve(),
    )
    INTERACTIVE_MODE: Mandatory[bool] = Mandatory(
        lambda: get_cli_arg("interactive"),
        lambda: False,
    )
    ACTIONS_SOURCE_FILE: Optional[Path] = Optional(
        lambda: maybe_path(get_cli_arg("workflow")),
        lambda: maybe_path(Env.GRANA_WORKFLOW_FILE),
    )
    WORKFLOW_LOADER_CLASS: Optional[LoaderClassType] = Optional(
        lambda: maybe_class_from_module(
            path_str=Env.GRANA_WORKFLOW_LOADER_SOURCE_FILE,
            class_name="WorkflowLoader",
            submodule_name="workflow.loader",
        )
    )
    ACTION_CLASSES_DIRECTORIES: Mandatory[t.List[str]] = Mandatory(
        lambda: Env.GRANA_ACTIONS_CLASS_DEFINITIONS_DIRECTORY,
    )
    DISPLAY_CLASS: Mandatory[DisplayClassType] = Mandatory(
        lambda: _maybe_display_class_by_name(get_cli_arg("display")),
        lambda: maybe_class_from_module(
            path_str=Env.GRANA_DISPLAY_SOURCE_FILE,
            class_name="Display",
            submodule_name="display",
        ),
        lambda: _maybe_display_class_by_name(Env.GRANA_DISPLAY_NAME),
        _get_default_display_class,
    )
    STRATEGY_CLASS: Mandatory[StrategyClassType] = Mandatory(
        _get_strategy_class_from_cli_arg,
        lambda: _maybe_strategy(Env.GRANA_STRATEGY_NAME),
        _get_default_strategy_class,
    )
    USE_COLOR: Mandatory[bool] = Mandatory(
        lambda: Env.GRANA_FORCE_COLOR,
        _isatty,
    )
    SHELL_INJECT_YIELD_FUNCTION: Mandatory[bool] = Mandatory(
        lambda: Env.GRANA_SHELL_INJECT_YIELD_FUNCTION,
    )
    STRICT_OUTCOMES_RENDERING: Mandatory[bool] = Mandatory(
        lambda: Env.GRANA_STRICT_OUTCOMES_RENDERING,
    )
    DEFAULT_SHELL_EXECUTABLE: Mandatory[str] = Mandatory(
        lambda: Env.GRANA_DEFAULT_SHELL_EXECUTABLE,
    )



"""
A strategy is an async-iterable object,
emitting actions one by one for further scheduling.
"""

from __future__ import annotations

import asyncio
import typing as t

import classlogging

from .actions.base import ActionBase, ActionSkip
from .actions.types import ActionStatus
from .workflow import Workflow

ST = t.TypeVar("ST", bound="BaseStrategy")

__all__ = [
    "BaseStrategy",
    "FreeStrategy",
    "SequentialStrategy",
    "LooseStrategy",
    "StrictStrategy",
    "StrictSequentialStrategy",
    "KNOWN_STRATEGIES",
]

KNOWN_STRATEGIES: t.Dict[str, t.Type[BaseStrategy]] = {}


class BaseStrategy(classlogging.LoggerMixin, t.AsyncIterable[ActionBase]):
    """Strategy abstract base"""

    NAME: str = ""
    STRICT: bool = False

    def __init__(self, workflow: Workflow) -> None:
        self._workflow = workflow

    def __aiter__(self: ST) -> ST:
        return self

    async def __anext__(self) -> ActionBase:
        raise NotImplementedError

    def __init_subclass__(cls, **kwargs):
        if KNOWN_STRATEGIES.setdefault(cls.NAME, cls) is not cls:
            raise NameError(
                f"Strategy named {cls.NAME!r} already exists. "
                f"Please specify another name for the {cls.__module__}.{cls.__name__}."
            )

    def _skip_action(self, action: ActionBase) -> None:
        try:
            action.skip()
        except ActionSkip:
            pass


class FreeStrategy(BaseStrategy):
    """Free execution (fully parallel)"""

    NAME = "free"

    def __init__(self, workflow: Workflow) -> None:
        super().__init__(workflow)
        self._unprocessed: t.List[ActionBase] = list(workflow.values())

    async def __anext__(self) -> ActionBase:
        if not self._unprocessed:
            raise StopAsyncIteration
        return self._unprocessed.pop(0)


class SequentialStrategy(FreeStrategy):
    """Sequential execution"""

    NAME = "sequential"

    def __init__(self, workflow: Workflow) -> None:
        super().__init__(workflow)
        self._current: t.Optional[ActionBase] = None

    async def __anext__(self) -> ActionBase:
        if self._current is not None:
            try:
                await self._current
            except Exception:
                if self.STRICT:
                    while True:
                        next_action = await super().__anext__()
                        self._skip_action(next_action)
        self._current = await super().__anext__()
        return self._current


class LooseStrategy(BaseStrategy):
    """Keep tracking dependencies states"""

    NAME = "loose"

    def __init__(self, workflow: Workflow) -> None:
        super().__init__(workflow)
        # Actions that have been emitted by the strategy and not finished yet
        self._active_actions_map: t.Dict[str, ActionBase] = {}
        # Just a structured mutable copy of the dependency map
        self._action_blockers: t.Dict[str, t.Set[str]] = {name: set(workflow[name].ancestors) for name in workflow}

    def _skip_action(self, action: ActionBase) -> None:
        super()._skip_action(action)
        self._active_actions_map.pop(action.name, None)

    def _get_maybe_next_action(self) -> t.Optional[ActionBase]:
        """Completely non-optimal (always scan all actions), but readable yet"""
        done_action_names: t.Set[str] = {action.name for action in self._workflow.values() if action.done()}
        # Copy into a list for further possible pop
        for maybe_next_action_name, maybe_next_action_blockers in list(self._action_blockers.items()):
            maybe_next_action_blockers -= done_action_names
            if not maybe_next_action_blockers:
                self.logger.debug(f"Action {maybe_next_action_name!r} is ready for scheduling")
                self._action_blockers.pop(maybe_next_action_name)
                next_action: ActionBase = self._workflow[maybe_next_action_name]
                self._active_actions_map[next_action.name] = next_action
                return next_action
        return None

    async def __anext__(self) -> ActionBase:
        while True:
            # Get an action and check whether to emit or to skip it
            next_action: ActionBase = await self._next_action()
            self.logger.debug(f"The next action is: {next_action}")
            for ancestor_name, ancestor_dependency in next_action.ancestors.items():
                ancestor: ActionBase = self._workflow[ancestor_name]
                if ancestor.status in (ActionStatus.FAILURE, ActionStatus.SKIPPED, ActionStatus.WARNING) and (
                    ancestor_dependency.strict or self.STRICT
                ):
                    self.logger.debug(f"Action {next_action} is qualified as skipped due to strict failure: {ancestor}")
                    self._skip_action(next_action)
                    break
            else:
                return next_action

    async def _next_action(self) -> ActionBase:
        # Do we have anything pending already?
        if maybe_next_action := self._get_maybe_next_action():
            return maybe_next_action
        # Await for any actions finished. Can't directly apply asyncio.wait to Action objects
        # since python 3.11's implementation requires too many methods from an awaitable object.
        while active_actions := list(self._active_actions_map.values()):
            await asyncio.wait(
                [action.get_future() for action in active_actions],
                return_when=asyncio.FIRST_COMPLETED,
            )
            for action in active_actions:  # type: ActionBase
                if action.done():
                    self.logger.debug(f"Action {action.name!r} execution finished")
                    del self._active_actions_map[action.name]
            # Maybe now?
            if maybe_next_action := self._get_maybe_next_action():
                return maybe_next_action
        raise StopAsyncIteration


class StrictStrategy(LooseStrategy):
    """Respect all dependencies, but force them strict"""

    STRICT = True
    NAME = "strict"


class StrictSequentialStrategy(SequentialStrategy):
    """Linear execution until first failure"""

    STRICT = True
    NAME = "strict-sequential"



"""CLI arguments"""

import functools
import typing as t

import click

__all__ = [
    "cliargs_receiver",
    "get_cli_arg",
]

_CLI_PARAMS: t.Dict[str, t.Any] = {}


def cliargs_receiver(func):  # pragma: no cover
    """Store CLI args in the _CLI_PARAMS container for further processing"""

    @functools.wraps(func)
    # pylint: disable=unused-argument
    def wrapped(ctx: click.Context, **kwargs):
        old_cli_params: t.Dict[str, t.Any] = _CLI_PARAMS.copy()
        current_ctx: t.Optional[click.Context] = ctx
        while current_ctx:
            for k, v in current_ctx.params.items():
                if k not in _CLI_PARAMS:
                    _CLI_PARAMS[k] = v
            current_ctx = current_ctx.parent
        try:
            return func()
        finally:
            # Restore CLI params container
            for k in list(_CLI_PARAMS):
                del _CLI_PARAMS[k]
            _CLI_PARAMS.update(old_cli_params)

    return click.pass_context(wrapped)


def get_cli_arg(name: str, *, valid_options: t.Optional[t.Iterable[str]] = None) -> t.Any:
    """Obtain previously registered CLI argument"""

    value: t.Any = _CLI_PARAMS.get(name)
    if valid_options is not None and value is not None and value not in valid_options:
        raise ValueError(
            f"Unrecognized value for the {name!r} argument: {value!r}. " f"Expected one of: {sorted(valid_options)}"
        )
    return value



"""Workflow-related module"""

from __future__ import annotations

import collections
import pathlib
import typing as t

from classlogging import LoggerMixin

from .actions.base import ActionBase, ActionDependency
from .exceptions import IntegrityError

__all__ = [
    "Workflow",
]


class Workflow(t.Dict[str, ActionBase], LoggerMixin):
    """Action relations map"""

    def __init__(
        self,
        actions_map: t.Dict[str, ActionBase],
        context: t.Optional[t.Dict[str, t.Any]] = None,
        source_file: t.Optional[pathlib.Path] = None,
    ) -> None:
        super().__init__(actions_map)
        self.source_file: t.Optional[pathlib.Path] = source_file
        self._entrypoints: t.Set[str] = set()
        self._tiers_sequence: t.List[t.List[ActionBase]] = []
        self._descendants_map: t.Dict[str, t.Dict[str, ActionDependency]] = collections.defaultdict(dict)
        self.context: t.Dict[str, t.Any] = context or {}
        # Check dependencies integrity
        self._establish_descendants()
        # Create order map to check all actions are reachable
        self._allocate_tiers()

    def get_metadata(self) -> t.Dict[str, t.Any]:
        """Obtain workflow metadata for further use in templating"""
        if self.source_file is None:
            return {}
        return {
            "source_file": self.source_file,
            "here": self.source_file.parent,
        }

    def _establish_descendants(self) -> None:
        missing_non_external_deps: t.Set[str] = set()
        for action in self.values():  # type: ActionBase
            for dependency_action_name, dependency in list(action.ancestors.items()):
                if dependency_action_name not in self:
                    if dependency.external:
                        # Get rid of missing external deps
                        action.ancestors.pop(dependency_action_name)
                    else:
                        missing_non_external_deps.add(dependency_action_name)
                    continue
                # Register symmetric descendant connection for further simplicity
                self._descendants_map[dependency_action_name][action.name] = dependency
            # Check if there are any dependencies after removal at all
            if not action.ancestors:
                self._entrypoints.add(action.name)
        if missing_non_external_deps:
            raise IntegrityError(f"Missing actions among dependencies: {sorted(missing_non_external_deps)}")
        # Check entrypoints presence
        if not self._entrypoints:
            raise IntegrityError("No entrypoints for the workflow")

    def _allocate_tiers(self) -> None:
        """Use Dijkstra algorithm to introduce partial order.
        The tier is a group of tasks, and it's called finished when all its tasks are processed.
        Tier #0 consists of entrypoints.
        Tier #N consists of all tasks requiring exactly N-1 preceding tiers to be finished.
        """
        step_tier: int = 0
        action_name_to_tier_mapping: t.Dict[str, int] = {}
        #
        current_tier_actions_names: t.Set[str] = self._entrypoints
        while True:
            next_tier_candidate_actions_names: t.Set[str] = set()
            for tier_action_name in current_tier_actions_names:
                tier_action: ActionBase = self[tier_action_name]
                if tier_action.name in action_name_to_tier_mapping:
                    continue
                action_name_to_tier_mapping[tier_action.name] = step_tier
                next_tier_candidate_actions_names |= set(self._descendants_map[tier_action_name])
            if not next_tier_candidate_actions_names:
                break
            step_tier += 1
            current_tier_actions_names = next_tier_candidate_actions_names
        self.logger.debug(f"Number of tiers: {step_tier + 1}")
        unreachable_action_names: t.Set[str] = {
            action.name for action in self.values() if action.name not in action_name_to_tier_mapping
        }
        if unreachable_action_names:
            raise IntegrityError(f"Unreachable actions found: {sorted(unreachable_action_names)}")
        self._tiers_sequence = [[] for _ in range(step_tier + 1)]
        for action_name, action in self.items():
            action_tier: int = action_name_to_tier_mapping[action_name]
            self._tiers_sequence[action_tier].append(action)

    def _iter_actions_by_tier(self) -> t.Generator[t.Tuple[int, ActionBase], None, None]:
        """Yield actions tier by tier"""
        for tier_num, tier_actions in enumerate(self._tiers_sequence):
            for action in tier_actions:
                yield tier_num, action

    def iterate_actions(self) -> t.Iterator[ActionBase]:
        """Iterate actions sorted natively"""
        return (action for _, action in self._iter_actions_by_tier())



"""Everything related to a default action interpretation"""

from __future__ import annotations

import asyncio
import base64
import collections
import enum
import re
import textwrap
import typing as t
from dataclasses import dataclass, fields

import classlogging

from .constants import ACTION_RESERVED_FIELD_NAMES
from .types import Stderr, OutcomeStorageType, ActionStatus
from ..display.types import DisplayEvent, DisplayEventName
from ..exceptions import ActionRunError

__all__ = [
    "ActionDependency",
    "ActionSeverity",
    "ActionBase",
    "ActionSkip",
    "ArgsBase",
    "EmissionScannerActionBase",
]


class ActionSkip(BaseException):
    """Stop executing action"""


class ActionSeverity(enum.Enum):
    """Action severity"""

    LOW = "low"
    NORMAL = "normal"


@dataclass
class ActionDependency:
    """Dependency info holder"""

    strict: bool = False
    external: bool = False


class ArgsMeta(type):
    """Metaclass for args containers that makes them all dataclasses"""

    def __new__(cls, name, bases, dct):
        sub_dataclass = dataclass(super().__new__(cls, name, bases, dct))
        reserved_names_collisions: t.Set[str] = {f.name for f in fields(sub_dataclass)} & ACTION_RESERVED_FIELD_NAMES
        if reserved_names_collisions:
            raise TypeError(f"Reserved names found in {name!r} class definition: {sorted(reserved_names_collisions)}")
        return sub_dataclass


@dataclass
class ArgsBase(metaclass=ArgsMeta):
    """Default empty args holder.
    Should be subclassed and then added to the `args` annotation of any action class."""


class ActionBase(classlogging.LoggerMixin):
    """Base class for all actions"""

    args: ArgsBase

    def __init__(
        self,
        name: str,
        args: ArgsBase = ArgsBase(),
        ancestors: t.Optional[t.Dict[str, ActionDependency]] = None,
        description: t.Optional[str] = None,
        selectable: bool = True,
        severity: ActionSeverity = ActionSeverity.NORMAL,
    ) -> None:
        self.name: str = name
        self.args: ArgsBase = args
        self.description: t.Optional[str] = description
        self.ancestors: t.Dict[str, ActionDependency] = ancestors or {}
        self.selectable: bool = selectable

        self._yielded_keys: OutcomeStorageType = {}
        self._status: ActionStatus = ActionStatus.PENDING
        self._enabled: bool = True
        # Do not create asyncio-related objects on constructing object to decouple from the event loop
        self._maybe_finish_flag: t.Optional[asyncio.Future] = None
        self._maybe_message_queue: t.Optional[asyncio.Queue[DisplayEvent]] = None
        self._running_task: t.Optional[asyncio.Task] = None
        self._severity: ActionSeverity = severity

    @property
    def enabled(self) -> bool:
        """Check whether the action has not been disabled"""
        return self._enabled

    def disable(self) -> None:
        """Marking the action as not planned for launch"""
        self.logger.info(f"Disabling {self}")
        if self._status != ActionStatus.PENDING:
            raise RuntimeError(f"Action {self.name} can't be disabled due to its status: {self._status!r}")
        self._enabled = False

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(name={self.name!r}, status={self._status.value})"

    def yield_outcome(self, key: str, value: t.Any) -> None:
        """Report outcome key"""
        self.logger.debug(f"Yielded a key: {key!r}")
        self._yielded_keys[key] = value

    def get_outcomes(self) -> OutcomeStorageType:
        """Report all registered outcomes"""
        return self._yielded_keys

    def get_future(self) -> asyncio.Future:
        """Return a Future object indicating the end of the action"""
        if self._maybe_finish_flag is None:
            self._maybe_finish_flag = asyncio.get_event_loop().create_future()
        return self._maybe_finish_flag

    @property
    def _event_queue(self) -> asyncio.Queue[DisplayEvent]:
        if self._maybe_message_queue is None:
            self._maybe_message_queue = asyncio.Queue()
        return self._maybe_message_queue

    @property
    def status(self) -> ActionStatus:
        """Public getter"""
        return self._status

    async def run(self) -> None:
        """Main entry to be implemented in subclasses"""
        raise NotImplementedError

    async def _run_with_log_context(self) -> None:
        self.logger.info(f"Running action: {self.name!r}")
        with self.logger.context(name=self.name):
            return await self.run()

    async def _await(self) -> None:
        fut = self.get_future()
        if fut.done():
            return fut.result()
        # Allocate asyncio task
        if self._running_task is None:
            self._running_task = asyncio.create_task(self._run_with_log_context())
            self._status = ActionStatus.RUNNING
        try:
            if (running_task_result := await self._running_task) is not None:
                self.logger.warning(f"Action {self.name!r} return type is {type(running_task_result)} (not NoneType)")
        except ActionSkip:
            pass
        except ActionRunError:
            raise
        except Exception as e:
            self._internal_fail(e)
            raise
        else:
            self._status = ActionStatus.SUCCESS
        if not fut.done():
            fut.set_result(None)

    def say(self, message: str) -> None:
        """Send a message to the display"""
        self._event_queue.put_nowait(DisplayEvent(DisplayEventName.ON_ACTION_MESSAGE, source=self, message=message))

    def skip(self) -> t.NoReturn:
        """Set status to SKIPPED and stop execution"""
        self._internal_skip()
        raise ActionSkip

    def _internal_skip(self) -> None:
        self._status = ActionStatus.SKIPPED
        self.get_future().set_result(None)
        self.logger.info(f"Action {self.name!r} skipped")

    def _internal_omit(self) -> None:
        self._status = ActionStatus.OMITTED
        self.get_future().set_result(None)
        self.logger.info(f"Action {self.name!r} omitted")

    def fail(self, message: str = "") -> t.NoReturn:
        """Set corresponding error message and raise an exception"""
        exception = ActionRunError(message)
        self._internal_fail(exception)
        raise exception

    def _internal_fail(self, exception: Exception) -> None:
        if not self.get_future().done():
            self._status = ActionStatus.FAILURE if self._severity == ActionSeverity.NORMAL else ActionStatus.WARNING
            self.logger.info(f"Action {self.name!r} failed: {repr(exception)}")
            self.get_future().set_exception(exception)

    async def read_messages(self) -> t.AsyncGenerator[DisplayEvent, None]:
        """Obtain all said messages sequentially"""
        while True:
            # Wait for either an event or action finish
            queue_getter = asyncio.create_task(self._event_queue.get())
            await asyncio.wait(
                [self.get_future(), queue_getter],
                return_when=asyncio.FIRST_COMPLETED,
            )
            if queue_getter.done():
                yield queue_getter.result()
            if self.done():
                # The action is done, so we should drain the queue.
                # Prevent queue from async get since then.
                queue_getter.cancel()
                while True:
                    try:
                        yield self._event_queue.get_nowait()
                    except asyncio.QueueEmpty:
                        break
                return

    def __await__(self) -> t.Generator[t.Any, None, None]:
        return self._await().__await__()  # pylint: disable=no-member

    def done(self) -> bool:
        """Indicate whether the action is over"""
        return self.get_future().done() or self._status in (ActionStatus.SKIPPED, ActionStatus.OMITTED)


# pylint: disable=abstract-method
class EmissionScannerActionBase(ActionBase):
    """Base class for stream-scanning actions"""

    _SERVICE_MESSAGES_SCAN_PATTERN: t.ClassVar[t.Pattern] = re.compile(
        r"""^
          (.*?)  # possible preceding content
          \#\#grana\[  # message prefix
            ([A-Za-z0-9+/=\- ]+)  # message itself
          ]\#\#  # message suffix
        $""",
        re.VERBOSE,
    )
    _SHELL_SERVICE_FUNCTIONS_DEFINITIONS: str = textwrap.dedent(
        r"""
            yield_outcome()(
              [ "$1" = "" ] && echo "Missing key (first argument)" && return 1
              [ "$3" != "" ] && echo "Too many arguments (expected 1 or 2)" && return 2
              command -v base64 >/dev/null || ( echo "Missing command: base64" && return 3 )
              _pipe()(
                encodedKey="$1"
                while read -r data; do
                  echo "##grana[yield-outcome-b64-chunk $encodedKey $data]##"
                done
                echo "##grana[yield-outcome-b64-end $encodedKey]##"
              )
              encodedKey=$( printf "%s" "$1" | base64 | tr -d '\n' )
              if [ "$2" = "" ]; then
                base64 </dev/stdin | _pipe "$encodedKey"
              else
                printf "%s" "$2" | base64 | _pipe "$encodedKey"
              fi
              return 0
            )
            skip(){
              echo "##grana[skip]##"
              exit 0
            }
        """
    ).lstrip()

    def __init__(self, *a, **kw) -> None:
        super().__init__(*a, **kw)
        self._outcomes_base64_chunks: t.Dict[str, t.List[str]] = collections.defaultdict(list)

    @classmethod
    def _decode_base64_string(cls, data: str) -> str:
        return base64.b64decode(data, validate=True).decode()

    def _process_service_message_expression(self, expression: str) -> None:
        try:
            expression_type, *encoded_args = expression.split()
            if expression_type == "skip":
                self.skip()
            elif expression_type == "yield-outcome-b64-chunk":
                key, value = encoded_args
                self._outcomes_base64_chunks[key].append(value)
            elif expression_type == "yield-outcome-b64-end":
                (encoded_key,) = encoded_args
                encoded_outcome_value: str = "".join(self._outcomes_base64_chunks.pop(encoded_key))
                self.yield_outcome(
                    key=self._decode_base64_string(encoded_key),
                    value=self._decode_base64_string(encoded_outcome_value),
                )
            else:
                raise ValueError(f"Unrecognized expression: {expression!r}")
        except ActionSkip:  # pylint: disable=try-except-raise
            raise
        except Exception:
            self.logger.warning("Failed while parsing system message", exc_info=True)

    def say(self, message: str) -> None:
        # Do not check stderr
        if isinstance(message, Stderr):
            super().say(message)
            return
        memorized_prefix: str = ""
        for line in message.splitlines():
            # `endswith` is a cheaper check than re.findall
            if not line.endswith("]##") or not (matches := self._SERVICE_MESSAGES_SCAN_PATTERN.findall(line)):
                super().say(memorized_prefix + line)
                memorized_prefix = ""
                continue
            for preceding_content, expression in matches:
                memorized_prefix += preceding_content
                self._process_service_message_expression(expression)
        # Do not forget to report system message prefix, if any
        if memorized_prefix:
            super().say(memorized_prefix)



"""Base interface class for all loaders"""

from __future__ import annotations

import collections
import contextlib
import importlib.metadata as im
import typing as t
from enum import Enum
from pathlib import Path

import dacite
from classlogging import LoggerMixin
from dacite.types import is_subclass
from packaging.requirements import Requirement

from ..actions.base import ActionBase, ArgsBase, ActionDependency, ActionSeverity
from ..actions.types import ObjectTemplate, qualify_string_as_potentially_renderable
from ..exceptions import LoadError, PackageRequirementsError
from ..strategy import KNOWN_STRATEGIES, BaseStrategy
from ..tools.concealment import represent_object_type
from ..tools.inspect import get_class_annotations
from ..workflow import Workflow

__all__ = [
    "AbstractBaseWorkflowLoader",
]


class TemplateIndifferentConfig(dacite.Config, LoggerMixin):
    """Configuration for initial workflow loading"""

    @classmethod
    def is_instance(cls, value: t.Any, type_: t.Type) -> bool:
        if isinstance(value, ObjectTemplate):
            cls.logger.info(f"Skipping type check for object template, where {type_!r} was expected")
            return True
        if is_subclass(type_, Enum):
            if isinstance(value, str) and qualify_string_as_potentially_renderable(value):
                cls.logger.info(f"Skipping type check for a renderable string, where {type_!r} was expected")
            else:
                # This is not subject to rendering: check enum right away
                type_(value)
            return True
        return super().is_instance(value, type_)


class AbstractBaseWorkflowLoader(LoggerMixin):
    """Loaders base class"""

    STATIC_ACTION_FACTORIES: t.Dict[str, t.Type[ActionBase]] = {}

    def __init__(self) -> None:
        self._actions: t.Dict[str, ActionBase] = {}
        self._raw_file_names_stack: t.List[str] = []
        self._resolved_file_paths_stack: t.List[Path] = []
        self._gathered_context: t.Dict[str, t.Any] = {}
        self._original_args_map: t.Dict[str, t.Dict[str, t.Any]] = {}
        self._action_type_counters: t.Dict[str, int] = collections.defaultdict(int)
        self._package_requirements: t.List[Requirement] = []
        self._explicit_strategy_class: t.Optional[t.Type[BaseStrategy]] = None

    @property
    def strategy_class(self) -> t.Optional[t.Type[BaseStrategy]]:
        """Return explicitly-set strategy class, if any"""
        return self._explicit_strategy_class

    def _register_action(self, action: ActionBase) -> None:
        if action.name in self._actions:
            self._throw(f"Action declared twice: {action.name!r}")
        self._actions[action.name] = action

    def _throw(self, message: str) -> t.NoReturn:
        """Raise loader exception from text"""
        raise LoadError(message=message, stack=self._raw_file_names_stack) from None

    def _internal_load(self, source_file: t.Union[str, Path]) -> None:
        """Load workflow partially from file (can be called recursively).
        :param source_file: either Path or string object pointing at a file"""
        with self._read_file(source_file) as file_data:
            self._internal_loads(file_data)

    def _get_context(self) -> Path:
        """Return active context directory for relative path resolution"""
        return self._resolved_file_paths_stack[-1].parent if self._resolved_file_paths_stack else Path()

    @contextlib.contextmanager
    def _read_file(self, source_file: t.Union[str, Path]) -> t.Iterator[bytes]:
        """Read file data"""
        source_file_raw_path: Path = Path(source_file)
        if not source_file_raw_path.is_absolute():
            source_file_raw_path = self._get_context() / source_file_raw_path
        source_resolved_file_path = source_file_raw_path.resolve()
        if source_resolved_file_path in self._resolved_file_paths_stack:
            self._throw("Cyclic load")
        self._raw_file_names_stack.append(str(source_file))
        self._resolved_file_paths_stack.append(source_resolved_file_path)
        self.logger.debug(f"Loading workflow file: {source_resolved_file_path}")
        try:
            if not source_resolved_file_path.is_file():
                self._throw(f"Workflow file not found: {source_resolved_file_path}")
            yield source_resolved_file_path.read_bytes()
        finally:
            self._raw_file_names_stack.pop()
            self._resolved_file_paths_stack.pop()

    def _internal_loads(self, data: t.Union[str, bytes]) -> None:
        """Load workflow partially from text (can be called recursively)"""
        raise NotImplementedError

    def _get_action_factory_by_type(self, action_type: str) -> t.Type[ActionBase]:
        if action_type not in self.STATIC_ACTION_FACTORIES:
            self._throw(f"Unknown dispatched type: {action_type}")
        return self.STATIC_ACTION_FACTORIES[action_type]

    def loads(self, data: t.Union[str, bytes]) -> Workflow:
        """Load workflow from text"""
        self._internal_loads(data=data)
        return Workflow(self._actions, context=self._gathered_context)

    def load(self, source_file: t.Union[str, Path]) -> Workflow:
        """Load workflow from file"""
        self._internal_load(source_file=source_file)
        return Workflow(self._actions, context=self._gathered_context, source_file=Path(source_file))

    def build_dependency_from_node(self, dep_node: t.Union[str, dict]) -> t.Tuple[str, ActionDependency]:
        """Unified method to process transform dependency source data"""
        dep_holder: ActionDependency = ActionDependency()
        if isinstance(dep_node, str):
            return dep_node, dep_holder
        if isinstance(dep_node, dict):
            unexpected_dep_keys: t.Set[str] = set(dep_node) - {"name", "strict", "external"}
            if unexpected_dep_keys:
                self._throw(f"Unrecognized dependency node keys: {sorted(unexpected_dep_keys)}")
            # Dependency name
            if "name" not in dep_node:
                self._throw(f"Name not specified for the dependency: {sorted(dep_node.items())}")
            dep_name: str = dep_node["name"]
            if not isinstance(dep_name, str):
                self._throw(f"Unrecognized dependency name type: {type(dep_name)!r} (expected a string)")
            if not dep_name:
                self._throw("Empty dependency name met")
            # Dependency 'strict' attr
            strict: bool = dep_node.get("strict", False)
            if not isinstance(strict, bool):
                self._throw(f"Unrecognized 'strict' attribute type: {type(strict)!r} (expected boolean)")
            dep_holder.strict = strict
            # Dependency 'external' attr
            external: bool = dep_node.get("external", False)
            if not isinstance(external, bool):
                self._throw(f"Unrecognized 'external' attribute type: {type(external)!r} (expected boolean)")
            dep_holder.external = external
            return dep_name, dep_holder
        self._throw(f"Unrecognized dependency node structure: {type(dep_node)!r} (expected a string or a dict)")

    def build_action_from_dict_data(self, node: dict) -> ActionBase:
        """Process a dictionary representing an action"""
        # Action type
        if "type" not in node:
            self._throw("'type' not specified for action")
        action_type: str = node.pop("type")
        action_class: t.Type[ActionBase] = self._get_action_factory_by_type(action_type)
        # Action name
        name: str
        if "name" in node:
            name = node.pop("name")
            if not isinstance(name, str):
                self._throw(f"Unexpected name type: {type(name)!r} (should be a string")
            if not name:
                self._throw("Action node name is empty")
        else:
            name = f"{action_type}-{self._action_type_counters[action_type]}"
        self._action_type_counters[action_type] += 1
        # Description
        description: t.Optional[str] = node.pop("description", None)
        if description is not None and not isinstance(description, str):
            self._throw(f"Unrecognized 'description' content type: {type(description)!r} (expected optional string)")
        # Dependencies
        deps_node: t.Union[str, t.List[t.Union[str, dict]]] = node.pop("expects", [])
        if not isinstance(deps_node, str) and not isinstance(deps_node, list):
            self._throw(f"Unrecognized 'expects' content type: {type(deps_node)!r} (expected a string or list)")
        if isinstance(deps_node, str):
            deps_node = [deps_node]
        dependencies: t.Dict[str, ActionDependency] = dict(
            self.build_dependency_from_node(dep_node) for dep_node in deps_node
        )
        # Selectable
        selectable: bool = node.pop("selectable", True)
        if not isinstance(selectable, bool):
            self._throw(f"Unrecognized 'selectable' content type: {type(selectable)!r} (expected a bool)")
        # Severity
        severity_str: str = node.pop("severity", ActionSeverity.NORMAL.value)
        if not isinstance(severity_str, str):
            self._throw(f"Unrecognized 'severity' content type: {type(severity_str)!r} (expected a string)")
        try:
            severity = ActionSeverity(severity_str)
        except ValueError:
            valid_severities: str = ", ".join(sorted(s.value for s in ActionSeverity))
            self._throw(f"Invalid severity: {severity_str!r} (expected one of: {valid_severities})")
        # Make action instance
        args_instance: ArgsBase = self._build_args_from_the_rest_of_the_dict_node(
            action_name=name,
            action_class=action_class,
            node=node,
        )
        action_instance: ActionBase = action_class(
            name=name,
            args=args_instance,
            description=description,
            ancestors=dependencies,
            selectable=selectable,
            severity=severity,
        )
        self._original_args_map[name] = node
        return action_instance

    def _build_args_from_the_rest_of_the_dict_node(
        self,
        action_name: str,
        action_class: t.Type[ActionBase],
        node: dict,
    ) -> ArgsBase:
        for mro_class in action_class.__mro__:
            if args_class := get_class_annotations(mro_class).get("args"):
                break
        else:
            self._throw(f"Couldn't find an `args` annotation for class {action_class.__name__}")
        try:
            return t.cast(
                ArgsBase,
                dacite.from_dict(
                    data_class=args_class,
                    data=node,
                    config=TemplateIndifferentConfig(
                        strict=True,
                        cast=[Path],
                    ),
                ),
            )
        except ValueError as e:
            self._throw(f"Action {action_name!r}: {e}")
        except dacite.MissingValueError as e:
            self._throw(f"Missing key for action {action_name!r}: {e.field_path!r}")
        except dacite.UnexpectedDataError as e:
            self._throw(f"Unrecognized keys for action {action_name!r}: {sorted(e.keys)}")
        except dacite.WrongTypeError as e:
            self._throw(
                f"Unrecognized {e.field_path!r} content type: {represent_object_type(e.value)}"
                f" (expected {e.field_type!r})"
            )

    def get_original_args_dict_for_action(self, action: ActionBase) -> dict:
        """Obtain dictionary representation of the action arguments as was initially loaded"""
        return self._original_args_map[action.name]

    def check_requirements(self) -> None:
        """Check that all loaded requirements are met"""
        failed_constrains: t.List[t.Tuple[t.Optional[str,], Requirement]] = []
        for package_constrain in self._package_requirements:
            self.logger.debug(f"Checking package requirement: {package_constrain}")
            try:
                installed_version: str = im.version(package_constrain.name)
            except im.PackageNotFoundError:
                failed_constrains.append((None, package_constrain))
            else:
                if installed_version not in package_constrain.specifier:
                    failed_constrains.append((installed_version, package_constrain))
        if failed_constrains:
            fails_info: t.List[str] = ["Following package requirements were not satisfied:"]
            for version, constrain in failed_constrains:
                required_version_details: str = f" of version {constrain.specifier}" if constrain.specifier else ""
                fails_info.append(f"    Requested {constrain.name}{required_version_details}, got {version}")
            raise PackageRequirementsError("\n".join(fails_info))

    def load_configuration_from_dict(self, configuration_dict: t.Dict[str, t.Any]) -> None:
        """Process configuration dictionary"""
        if not isinstance(configuration_dict, dict):
            self._throw(f"'configuration' contents should be a dict (got {type(configuration_dict)!r})")
        allowed_cfg_keys: t.Set[str] = {"requires_packages", "strategy"}
        if bad_cfg_keys := set(configuration_dict) - allowed_cfg_keys:
            self._throw(
                f"Unrecognized configuration keys: {sorted(bad_cfg_keys)}"
                f" (expected some of: {sorted(allowed_cfg_keys)})"
            )
        if "requires_packages" in configuration_dict:
            required_packages_list: t.Union[t.List[str], str] = configuration_dict["requires_packages"]
            if isinstance(required_packages_list, str):
                required_packages_list = [required_packages_list]
            elif not isinstance(required_packages_list, list):
                self._throw(
                    f"'configuration.requires_packages' contents should be "
                    f"a list of strings or a string (got {type(required_packages_list)!r})"
                )
            for package_constrain_str in required_packages_list:
                if not isinstance(package_constrain_str, str):
                    self._throw(f"Unexpected package constrain: {package_constrain_str!r} (expected a string)")
                package_constrain = Requirement(package_constrain_str)
                self.logger.debug(f"Defining package constrain: {package_constrain!r}")
                self._package_requirements.append(package_constrain)
        if "strategy" in configuration_dict:
            strategy_value: str = configuration_dict["strategy"]
            if strategy_value not in KNOWN_STRATEGIES:
                self._throw(f"Unexpected strategy: {strategy_value!r}")
            self._explicit_strategy_class = KNOWN_STRATEGIES[strategy_value]



"""Common loader utilities"""

import io
import typing as t
from pathlib import Path

from .base import AbstractBaseWorkflowLoader
from .default import DefaultYAMLWorkflowLoader
from ..exceptions import SourceError

__all__ = [
    "get_default_loader_class_for_source",
]

STREAM_DEFAULT_LOADER: t.Type[AbstractBaseWorkflowLoader] = DefaultYAMLWorkflowLoader
SUFFIX_TO_LOADER_MAP: t.Dict[str, t.Type[AbstractBaseWorkflowLoader]] = {
    ".yml": DefaultYAMLWorkflowLoader,
    ".yaml": DefaultYAMLWorkflowLoader,
}


def get_default_loader_class_for_source(
    source: t.Union[str, Path, io.TextIOBase],
) -> t.Type[AbstractBaseWorkflowLoader]:
    """Return loader class based on file stats"""
    if isinstance(source, io.TextIOBase):
        return STREAM_DEFAULT_LOADER
    source_path: Path = Path(source)
    if (loader_class := SUFFIX_TO_LOADER_MAP.get(source_path.suffix)) is None:
        raise SourceError(f"Unrecognized source: {source_path}")
    return loader_class



"""Lazy-loaded constants helpers"""

import hashlib
import os
import sys
import types
import typing as t
from contextlib import contextmanager
from importlib.machinery import ModuleSpec
from importlib.util import (
    spec_from_file_location,
    module_from_spec,
)
from pathlib import Path
from types import ModuleType

from ..environment import Env
from ...exceptions import SourceError

__all__ = [
    "Optional",
    "Mandatory",
    "maybe_path",
    "maybe_class_from_module",
]

VT = t.TypeVar("VT")
GetterType = t.Callable[[], t.Optional[VT]]

EXTERNALS_MODULES_PACKAGE: str = "grana.external"


@contextmanager
def add_sys_paths(*paths: str) -> t.Iterator[None]:
    """Temporarily add paths to sys.path"""
    normalized_paths: t.List[str] = [os.path.expanduser(os.path.abspath(path)) for path in paths]
    for path in normalized_paths:
        sys.path.insert(0, path)
    try:
        yield
    finally:
        for path in normalized_paths:
            sys.path.remove(path)


def load_external_module(source: Path, submodule_name: t.Optional[str] = None) -> ModuleType:
    """Load an external module"""
    if not source.is_file():
        raise SourceError(f"Missing source module: {source}")
    if submodule_name is None:
        submodule_name = hashlib.md5(str(source).encode()).hexdigest()  # nosec  # pragma: no cover
    module_name: str = f"{EXTERNALS_MODULES_PACKAGE}.{submodule_name}"
    module_spec: t.Optional[ModuleSpec] = spec_from_file_location(
        name=module_name,
        location=source,
    )
    if module_spec is None:
        raise SourceError(f"Can't read module spec from source: {source}")
    module: ModuleType = module_from_spec(module_spec)
    with add_sys_paths(*Env.GRANA_EXTERNAL_MODULES_PATHS):
        module_spec.loader.exec_module(module)  # type: ignore
    sys.modules[module_name] = module
    return module


class Optional(t.Generic[VT]):
    """Optional lazy variable"""

    def __init__(self, *getters: GetterType) -> None:
        self._getters: t.Tuple[GetterType, ...] = getters
        self._name: str = ""

    def __set_name__(self, owner: type, name: str) -> None:
        self._name = name

    def __get__(self, instance: t.Any, owner: type) -> t.Optional[VT]:
        getter_result: t.Optional[VT] = None
        for getter in self._getters:
            if (getter_result := getter()) is not None:
                break
        return getter_result


class Mandatory(Optional, t.Generic[VT]):
    """Mandatory lazy variable"""

    def __get__(self, instance: t.Any, owner: type) -> VT:
        result: t.Optional[VT] = super().__get__(instance, owner)
        if result is None:
            raise ValueError(f"{self._name!r} getters failed")
        return result


def maybe_path(path_str: str) -> t.Optional[Path]:
    """Transform a string into an optional path"""
    return Path(path_str) if path_str else None


def maybe_class_from_module(path_str: str, class_name: str, submodule_name: t.Optional[str] = None) -> t.Optional[type]:
    """Get a class from an external module, if given"""
    if (source_path := maybe_path(path_str)) is None:
        return None
    module: types.ModuleType = load_external_module(source_path, submodule_name)
    if not hasattr(module, class_name):
        raise AttributeError(f"External module contains no class {class_name!r} in {path_str!r}")
    return getattr(module, class_name)



"""Separate module for subflow action"""

import functools
import typing as t
from collections.abc import Mapping, MutableMapping
from dataclasses import field
from pathlib import Path

from ..base import ArgsBase, ActionBase
from ..types import (
    NamedMessageSource,
    RenamedMessageSource,
)
from ...display.types import DisplayEvent, DisplayEventName
from ...exceptions import ExecutionFailed
from ...rendering.containers import OutcomeDict

__all__ = [
    "SubflowAction",
]

ContextType = t.Dict[str, t.Any]


class SubflowArgs(ArgsBase):
    """Arguments applied to the subflow action."""

    path: Path
    context: t.Dict[str, t.Any] = field(default_factory=dict)  # pylint: disable=invalid-field-call


class SubflowAction(ActionBase):
    """Executes an independent workflow and passes the display events to the original runner."""

    args: SubflowArgs

    async def run(self) -> None:
        from ...runner import Runner  # pylint: disable=import-outside-toplevel,cyclic-import

        action: SubflowAction = self

        @functools.lru_cache()
        def _compose_source(origin: NamedMessageSource) -> NamedMessageSource:
            return RenamedMessageSource(name=f"{self.name}/{origin.name}", origin=origin)

        def _resend_event_via_action(event: DisplayEvent) -> None:
            # These events shall not pass to the parent runner
            if event.name in (
                DisplayEventName.ON_RUNNER_FINISH,  # Triggers final status output
                DisplayEventName.ON_PLAN_INTERACTION,  # Pauses the execution
            ):
                event.future.set_result(None)  # Unlock the execution and continue
            elif event.name == DisplayEventName.ON_RUNNER_START:
                event.kwargs["children"] = map(_compose_source, event.kwargs["children"])
                self._event_queue.put_nowait(event)  # Pass modified event
            elif event.name in (
                DisplayEventName.ON_ACTION_START,
                DisplayEventName.ON_ACTION_FINISH,
                DisplayEventName.ON_ACTION_MESSAGE,
                DisplayEventName.ON_ACTION_ERROR,
            ):
                event.kwargs["source"] = _compose_source(event.kwargs["source"])
                self._event_queue.put_nowait(event)  # Pass modified event
            else:
                # Just in case we add some event types later and not specify behaviour here
                raise ValueError(f"Unknown event name: {event.name!r}")  # pragma: no cover

        class SubflowRunner(Runner):
            """A runner that intercepts and filters out events"""

            async def _process_display_events(self) -> None:
                while True:
                    event: DisplayEvent = await self._events_flow.get()
                    _resend_event_via_action(event)

            @classmethod
            def _deep_update_context(cls, receiver: ContextType, source: Mapping, path: str) -> ContextType:
                """Apply changes to the context"""
                for source_key, source_value in source.items():
                    sub_path: str = f"{path}.{source_key}" if path else source_key
                    if source_key not in receiver:
                        cls.logger.debug(f"Adding context: {sub_path}")
                        receiver[source_key] = source_value
                    elif isinstance(source_value, Mapping) and isinstance(receiver[source_key], MutableMapping):
                        cls.logger.trace(f"Merging context: {sub_path}")
                        receiver[source_key] = cls._deep_update_context(receiver[source_key], source_value, sub_path)
                    else:
                        cls.logger.debug(f"Rewriting context: {sub_path}")
                        receiver[source_key] = source_value
                return receiver

            async def run_async(self) -> None:
                self.workflow.context = self._deep_update_context(
                    receiver=self.workflow.context,
                    source=action.args.context,
                    path="",
                )
                try:
                    return await super().run_async()
                finally:
                    for sub_action_name, sub_action_outcomes in self._outcomes.items():
                        action.yield_outcome(sub_action_name, OutcomeDict(sub_action_outcomes))

        runner = SubflowRunner(source=self.args.path)
        try:
            await runner.run_async()
        except ExecutionFailed:
            self.fail()



"""Separate module for shell-related action"""

import asyncio
import os
import typing as t

from async_shell import Shell, ShellResult

from ..base import ArgsBase, EmissionScannerActionBase
from ..types import Stderr
from ...config.constants import C

__all__ = [
    "ShellArgs",
    "ShellAction",
]


class ShellArgs(ArgsBase):
    """Args for shell-related actions"""

    command: t.Optional[str] = None
    file: t.Optional[str] = None
    environment: t.Optional[t.Dict[str, str]] = None
    cwd: t.Optional[str] = None
    executable: t.Optional[str] = None

    def __post_init__(self) -> None:
        if self.command is None and self.file is None:
            raise ValueError("Neither command nor file specified")
        if self.command is not None and self.file is not None:
            raise ValueError("Both command and file specified")


class ShellAction(EmissionScannerActionBase):
    """Shell commands handler"""

    args: ShellArgs

    async def _read_stdout(self, shell_process: Shell) -> None:
        async for line in shell_process.read_stdout():
            self.say(line)

    async def _read_stderr(self, shell_process: Shell) -> None:
        async for line in shell_process.read_stderr():
            self.say(Stderr(line))

    async def _create_shell(self) -> Shell:
        command: str = self.args.command or f"source '{self.args.file}'"
        if C.SHELL_INJECT_YIELD_FUNCTION:
            command = f"{self._SHELL_SERVICE_FUNCTIONS_DEFINITIONS}\n{command}"
        environment: t.Optional[t.Dict[str, str]] = None
        if self.args.environment is not None:
            environment = os.environ.copy()
            environment.update(self.args.environment)
        return Shell(
            command=command,
            environment=environment,
            cwd=self.args.cwd,
            executable=self.args.executable or C.DEFAULT_SHELL_EXECUTABLE,
        )

    async def run(self) -> None:
        async with await self._create_shell() as shell_process:
            tasks: t.List[asyncio.Task] = [
                asyncio.create_task(self._read_stdout(shell_process)),
                asyncio.create_task(self._read_stderr(shell_process)),
            ]
            # Wait for all tasks to complete
            await asyncio.wait(tasks)
            # Check exceptions
            await asyncio.gather(*tasks)
            result: ShellResult = await shell_process
            if result.code:
                self.fail(f"Exit code: {result.code}")
